from __future__ import absolute_import
from flask import url_for, render_template, redirect, abort, flash, request,\
        Blueprint, current_app
from flask.ext.login import login_required, current_user
from flask.ext.wtf import Form
import six
from six.moves import map
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from wtforms.fields import StringField, SubmitField, HiddenField, SelectField,\
        Label
from wtforms.validators import InputRequired, AnyOf, NumberRange

from ..models import db
from ..auth import PermissionType
from ..auth.models import Division, Permission, Entity
from ..util import jsonify


blueprint = Blueprint('divisions', __name__)


@blueprint.route('/')
@login_required
def list_divisions():
    """Show a page listing all divisions.

    Accesible only to administrators.
    """
    if current_user.admin:
        return render_template('divisions.html',
                divisions=Division.query.all())
    if current_user.has_permission(PermissionType.admin):
        admin_permissions = current_user.permissions.filter_by(
                permission=PermissionType.admin).values(Permission.division_id)
        admin_permissions = map(lambda x: x[0], admin_permissions)
        divisions = db.session.query(Division).\
                filter(Division.id.in_(admin_permissions))
        return render_template('divisions.html', divisions=divisions)
    return abort(403)


class AddDivisionForm(Form):
    name = StringField(u'Division Name', validators=[InputRequired()])
    submit = SubmitField(u'Create Division')


@blueprint.route('/add/', methods=['GET', 'POST'])
@login_required
def add_division():
    """Present a form for adding a division and also process that form.

    Only accesible to adminstrators.
    """
    if not current_user.admin:
        return abort(403)
    form = AddDivisionForm()
    if form.validate_on_submit():
        division = Division(form.name.data)
        db.session.add(division)
        db.session.commit()
        return redirect(url_for('.get_division_details', division_id=division.id))
    return render_template('form.html', form=form)


class ChangeEntity(Form):
    form_id = HiddenField(default='entity')
    id_ = HiddenField()
    name = StringField()
    permission = HiddenField(validators=[AnyOf(list(PermissionType.values()))])
    action = HiddenField(validators=[AnyOf(('add', 'delete'))])


#: List of tuples enumerating attributes that can be transformed/linked.
#: Mainly used as the choices argument to :py:class:`~.SelectField`
transformer_choices = [
    ('', u''),
    ('kill_timestamp', u'Kill Timestamp'),
    ('pilot', u'Pilot'),
    ('corporation', u'Corporation'),
    ('alliance', u'Alliance'),
    ('system', u'Solar System'),
    ('constellation', u'Constellation'),
    ('region', u'Region'),
    ('ship_type', u'Ship'),
    ('payout', u'Payout'),
    ('status', u'Request Status'),
]


class ChangeTransformer(Form):
    form_id = HiddenField(default='transformer')
    attribute = SelectField(u'Attribute', choices=transformer_choices)
    transformer = SelectField(u'Transformer', choices=[])


def transformer_choices(attr):
    """Conveniece function for generating a list of transformer option tuples.

    :param attr str: the name of the attribute to make a list for.
    :return: A list of tuples suitable for the choices argument of\
        :py:class:`StringField`
    :rtype: list
    """
    default_transformers = [
        ('none', 'None'),
    ]
    choices = default_transformers
    if attr in current_app.url_transformers:
        for transformer in current_app.url_transformers[attr]:
            choices.append((transformer, transformer))
    return choices


@blueprint.route('/<int:division_id>/', methods=['GET'])
@login_required
def get_division_details(division_id=None, division=None):
    """Generate a page showing the details of a division.

    Shows which groups and individuals have been granted permissions to each
    division.

    Only accesible to administrators.

    :param int division_id: The ID number of the division
    """
    if division is None:
        division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    if request.is_json or request.is_xhr:
        # Non-default way of encoding the division to include all the info in a
        # normal response
        entities = {}
        for perm in PermissionType.all:
            members = []
            for member in [p.entity for p in division.permissions[perm]]:
                member_info = {
                    'id': member.id,
                    'name': member.name,
                    'source': member.authmethod,
                }
                if hasattr(member, 'users'):
                    member_info['length'] = len(member.users)
                members.append(member_info)
            entities[perm.name] = members
        return jsonify(name=division.name, entities=entities, id=division.id)
    return render_template(
            'division_detail.html',
            division=division,
            entity_form=ChangeEntity(formdata=None),
            transformer_form=ChangeTransformer(formdata=None),
    )


def _modify_division_entity(division):
    """Handle POST requests for adding/removing entities form a Division."""
    form = ChangeEntity()
    if form.validate():
        entity = None
        if form.id_.data != '':
            current_app.logger.debug("Looking up entity by ID: {}".format(
                form.id_.data))
            entity = Entity.query.get(form.id_.data)
            if entity is None:
                flash(u"No entity with ID #{}.".format(form.id_.data),
                        u'error')
        else:
            current_app.logger.debug(u"Looking up entity by name: '{}'"\
                    .format(form.name.data))
            try:
                entity = Entity.query.filter_by(
                        name=form.name.data).one()
            except NoResultFound:
                flash(u"No entities with the name '{}' found.".
                        format(form.name.data), category=u'error')
            except MultipleResultsFound:
                flash(u"Multiple entities with the name '{}' found.".
                        format(form.name.data), category=u'error')
            else:
                current_app.logger.debug("entity lookup success")
        if entity is None:
            return get_division_details(division=division), 404, None
        # The entity has been found, create the query for the requested
        # Permission.
        permission_type = PermissionType.from_string(
                form.permission.data)
        permission_query = Permission.query.filter_by(
                division=division,
                entity=entity,
                permission=permission_type)
        # The response for both add and delete actions depends on wether the
        # Permission is found, so look it up first.
        try:
            permission = permission_query.one()
        except NoResultFound:
            if form.action.data == 'add':
                db.session.add(
                    Permission(division, permission_type, entity))
                flash(u"'{}' is now a {}.".format(entity,
                        permission_type.description.lower()), u"info")
            elif form.action.data == 'delete':
                flash(u"{} is not a {}.".format(entity,
                    permission_type.description.lower()), u"warning")
        else:
            if form.action.data == 'delete':
                permission_query.delete()
                flash(u"'{}' is no longer a {}.".format(entity,
                        permission_type.description.lower()), u"info")
            elif form.action.data == 'add':
                flash(u"'{}' is now a {}.".format(entity,
                        permission_type.description.lower()), u"info")
        db.session.commit()
    else:
        for field_name, errors in six.iteritems(form.errors):
            errors = u", ".join(errors)
            flash(u"Errors for {}: {}".format(field_name, errors), u'error')
        current_app.logger.info("Malformed entity permission POST: {}".format(
                form.errors))
    return get_division_details(division=division)


def _modify_division_transformer(division):
    """Handle POST requests for changing the Transformers for a Division."""
    form = ChangeTransformer()
    # Set the form's choices
    attr = form.attribute.data
    form.transformer.choices = transformer_choices(attr)
    # Check form and go from there
    if form.validate():
        name = form.transformer.data
        if name == 'none':
            division.transformers[attr] = None
        else:
            # Get the specific map of transformers for the attribute
            attr_transformers = current_app.url_transformers[attr]
            # Get the new transformer
            division.transformers[attr] = attr_transformers[name]
            # Explicitly add the TransformerRef to the session
            db.session.add(division.division_transformers[attr])
        db.session.commit()
        flash(u"'{}' transformer set to '{}'.".format(attr, name),
                u'message')
    else:
        for field_name, errors in six.iteritems(form.errors):
            errors = u", ".join(errors)
            flash(u"Errors for {}: {}".format(field_name, errors), u'error')
        current_app.logger.info("Malformed division transformer POST: {}".
                format(form.errors))
    return get_division_details(division=division)


@blueprint.route('/<int:division_id>/', methods=['POST'])
@login_required
def modify_division(division_id):
    """Dispatches modification requests to the specialized view function for
    that operation.
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    form_id = request.form.get('form_id')
    if form_id == 'entity':
        return _modify_division_entity(division)
    elif form_id == 'transformer':
        return _modify_division_transformer(division)
    else:
        current_app.logger.warn("Invalid division modification POST: {}"
                .format(request.form))
        abort(400)


@blueprint.route('/<int:division_id>/transformers/')
@blueprint.route('/<int:division_id>/transformers/<attribute>/')
@login_required
def list_transformers(division_id, attribute=None):
    """API method to get a list of transformers for a division.

    :param division_id int: the ID of the division to look up
    :param attribute str: a specific attribute to look up. Optional.
    :return: JSON
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    if attribute is None:
        attrs = six.iterkeys(current_app.url_transformers)
    else:
        attrs = (attribute,)
    choices = {}
    for attr in attrs:
        raw_choices = transformer_choices(attr)
        current = division.transformers.get(attr, None)
        if current is not None:
            choices[attr] = \
                    [(c[0], c[1], c[1] == current.name) for c in raw_choices]
        else:
            choices[attr] = \
                    [(c[0], c[1], False) for c in raw_choices]
    return jsonify(choices)



from __future__ import absolute_import
import hashlib

from flask import flash, url_for, redirect, abort, request
import six
from sqlalchemy.orm.exc import NoResultFound
from wtforms.fields import StringField, PasswordField, HiddenField, SubmitField
from wtforms.validators import InputRequired

from .. import db, requests_session
from ..util import ensure_unicode
from . import AuthMethod, AuthForm
from .models import User, Group, Pilot


class TestLoginForm(AuthForm):
    username = StringField(u'Username', validators=[InputRequired()])
    password = PasswordField(u'Password', validators=[InputRequired()])
    submit = SubmitField(u'Log In')


class TestAuth(AuthMethod):
    def __init__(self, api_key=None, **kwargs):
        self.api_key = api_key
        if 'name' not in kwargs:
            kwargs['name'] = u'Test Auth'
        super(TestAuth, self).__init__(**kwargs)

    def form(self):
        return TestLoginForm

    def login(self, form):
        sha = hashlib.sha1()
        password = form.password.data
        if isinstance(password, six.text_type):
            password = password.decode('utf-8')
        sha.update(password)
        params = {
                'user': form.username.data,
                'pass': sha.hexdigest()
                }
        response = requests_session.get(
                'https://auth.pleaseignore.com/api/1.0/login',
                params=params)
        json = response.json()
        # JSON is unicode, by definition
        if json[u'auth'] == u'failed':
            if json[u'error'] == u'none':
                flash(u"User '{}' not found.".format(form.username.data),
                        category=u'error')
            elif json[u'error'] == u'multiple':
                flash(u"Multiple users found.", category=u'error')
            elif json[u'error'] == u'password':
                flash(u"Incorrect password.", category=u'error')
            return redirect(url_for('login.login'))
        elif json[u'auth'] == u'ok':
            try:
                user = TestUser.query.filter_by(auth_id=json[u'id'],
                        authmethod=self.name).one()
            except NoResultFound:
                # Create new User
                user = TestUser(json[u'username'], json[u'id'], self.name)
                db.session.add(user)
            # Update values from Auth
            user.admin = json[u'superuser'] or json[u'staff'] or \
                    json[u'username'] in self.admins
            # Sync up group values
            for group in json[u'groups']:
                try:
                    db_group = TestGroup.query.\
                            filter_by(auth_id=group[u'id'],
                                    authmethod=self.name).one()
                except NoResultFound:
                    db_group = TestGroup(group[u'name'], group[u'id'],
                            self.name)
                    db.session.add(db_group)
                user.groups.add(db_group)
                # TODO: Remove old groups
            # Sync pilot associations
            pilot = Pilot.query.get(json[u'primarycharacter'][u'id'])
            if not pilot:
                pilot = Pilot(user, json[u'primarycharacter'][u'name'],
                        json[u'primarycharacter'][u'id'])
            pilot.user = user
            # Getting all pilots requires an Auth API key.
            if self.api_key:
                resp_user = requests_session.get(
                        'https://auth.pleaseignore.com/api/1.0/user', params=
                        {
                            'userid': user.id,
                            'apikey': self.api_key
                        })
                if resp_user.status_code == 200:
                    for char in resp_user.json()[u'characters']:
                        try:
                            pilot = Pilot.query.get(char[u'id'])
                        except NoResultFound:
                            pilot = Pilot(user, char[u'name'], char[u'id'])
                        else:
                            pilot.user = user
            # All done
            db.session.commit()
            self.login_user(user)
            return redirect(request.args.get('next') or url_for('index'))
        else:
            # Not sure what you did to get here, but somehow Auth has returned
            # an invalid response.
            abort(403)

    def list_groups(self, user=None):
        """Return a list of groups descriptors.

        If user is None, return _all_ groups. Otherwise, return the groups a
        member is part of.
        """
        if user is None:
            response = requests_session.get(
                    'https://auth.pleaseignore.com/api/1.0/info',
                    params={'request': 'groups'})
            # TODO Handle possible errors
            groups = set()
            for group in response.json():
                group_tuple = (group[u'name'], cls.__name__)
                groups.add(group_tuple)
            return groups
        else:
            # NOTE: THis might not be a secure/proper check. Test it.
            if user.authmethod() != cls:
                # TODO: Raise an exception here, this is the wrong authmethod
                # for this user.
                return None
            # TODO: Needs an Auth API key passed in somehow
            response = requests_session.get(
                    'https://auth.pleaseignore.com/api/1.0/user',
                    params={'userid': user.auth_id(), 'apikey': self.apikey})
            groups = set()
            for group in response.json()[u'groups']:
                group_tuple = (group[u'name'], cls.__name__)
                groups.add(group_tuple)
            return groups


class TestUser(User):
    id = db.Column(db.Integer, db.ForeignKey('user.id'), primary_key=True)
    auth_id = db.Column(db.Integer, nullable=False, index=True)

    def __init__(self, username, auth_id, authmethod, groups=None, **kwargs):
        self.name = ensure_unicode(username)
        self.auth_id = auth_id
        self.authmethod = ensure_unicode(authmethod)


class TestGroup(Group):
    id = db.Column(db.Integer, db.ForeignKey('group.id'), primary_key=True)
    auth_id = db.Column(db.Integer, nullable=False, index=True)
    description = db.Column(db.Text)

    def __init__(self, name, auth_id, authmethod):
        self.name = ensure_unicode(name)
        self.auth_id = auth_id
        self.authmethod = ensure_unicode(authmethod)



from .classproperty import classproperty
from .enum import DeclEnum
from .jsonify import jsonify
from .models import AutoID, Timestamped, AutoName
from .request import AcceptRequest
from .sqlstats import DB_STATS
from .unistr import unistr, ensure_unicode
from .urlparse import urlparse, urlunparse
from .utc import utc
from .xmlify import xmlify



import flask 


def jsonify(*args, **kwargs):
    messages = [{'message': m, 'category':c} for c, m in
            flask.get_flashed_messages(with_categories=True)]
    return flask.jsonify(*args, flashed_messages=messages, **kwargs)



from __future__ import unicode_literals
from flask import Request


class AcceptRequest(Request):

    _json_mimetypes = ['application/json',]

    _html_mimetypes = ['text/html', 'application/xhtml+xml']

    _xml_mimetypes = ['application/xml', 'text/xml']

    _rss_mimetypes = ['application/rss+xml', 'application/rdf+xml']

    @property
    def _known_mimetypes(self):
        return self._json_mimetypes + \
               self._html_mimetypes + \
               self._xml_mimetypes + \
               self._rss_mimetypes

    @property
    def is_json(self):
        if 'fmt' in self.values:
            return self.values['fmt'] == 'json'
        return self.accept_mimetypes.best_match(self._known_mimetypes) in \
            self._json_mimetypes

    @property
    def is_xml(self):
        if 'fmt' in self.values:
            return self.values['fmt'] == 'xml'
        return self.accept_mimetypes.best_match(self._known_mimetypes) in \
            self._xml_mimetypes

    @property
    def is_rss(self):
        if self.path.endswith('rss.xml'):
            return True
        if 'fmt' in self.values:
            return self.values['fmt'] == 'rss'
        return self.accept_mimetypes.best_match(self._known_mimetypes) in \
            self._rss_mimetypes



from __future__ import absolute_import
from base64 import urlsafe_b64encode
from itertools import groupby
import os
import six
from six.moves import filter
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.collections import attribute_mapped_collection, collection

from .. import db
from . import AuthMethod, PermissionType
from ..util import AutoID, Timestamped, AutoName, unistr, ensure_unicode
from ..models import Action, Modifier, Request


if six.PY3:
    unicode = str


users_groups = db.Table('users_groups', db.Model.metadata,
        db.Column('user_id', db.Integer, db.ForeignKey('user.id')),
        db.Column('group_id', db.Integer, db.ForeignKey('group.id')))


@unistr
class Entity(db.Model, AutoID, AutoName):
    """Private class for shared functionality between :py:class:`User` and
    :py:class:`Group`.

    This class defines a number of helper methods used indirectly by User and
    Group subclasses such as automatically defining the table name and mapper
    arguments.

    You should `not` inherit fomr this class directly, and should instead
    inherit from either :py:class:`User` or :py:class:`Group`.
    """

    #: The name of the entity. Usually a nickname.
    name = db.Column(db.String(100, convert_unicode=True), nullable=False)

    #: Polymorphic discriminator column.
    type_ = db.Column(db.String(50, convert_unicode=True))

    #: :py:class:`Permission`\s associated specifically with this entity.
    entity_permissions = db.relationship('Permission', collection_class=set,
            back_populates='entity', lazy='dynamic')

    #: The name of the :py:class:`AuthMethod` for this entity.
    authmethod = db.Column(db.String(50, convert_unicode=True), nullable=False)

    @declared_attr
    def __mapper_args__(cls):
        """SQLAlchemy late-binding attribute to set mapper arguments.

        Obviates subclasses from having to specify polymorphic identities.
        """
        cls_name = unicode(cls.__name__)
        args = {'polymorphic_identity': cls_name}
        if cls_name == u'Entity':
            args['polymorphic_on'] = cls.type_
        return args

    def __init__(self, name, authmethod, **kwargs):
        self.name = ensure_unicode(name)
        self.authmethod = ensure_unicode(authmethod)
        super(Entity, self).__init__(**kwargs)

    def __repr__(self):
        return "{x.__class__.__name__}('{x.name}')".format(x=self)

    def __unicode__(self):
        return u"{x.name}".format(x=self)

    def has_permission(self, permissions, division_or_request=None):
        """Returns if this entity has been granted a permission in a division.

        If ``division`` is ``None``, this method checks if this group has the
        given permission in `any` division.

        :param permissions: The series of permissions to check
        :type permissions: iterable
        :param division_or_request: The division to check. May also be ``None``
            or an SRP request.
        :type division: :py:class:`Division` or :py:class:`~.models.Request`
        :rtype bool:
        """
        if permissions in PermissionType.all:
            permissions = (permissions,)
        # admin permission includes the reviewer and payer privileges
        if PermissionType.admin not in permissions and \
                PermissionType.elevated.issuperset(permissions):
            if self.has_permission(PermissionType.admin, division_or_request):
                return True
        perms = self.permissions.filter(Permission.permission.in_(permissions))
        if division_or_request is not None:
            # requests have a 'division' attribute, so we check for that
            if hasattr(division_or_request, 'division'):
                division = division_or_request.division
            else:
                division = division_or_request
            perms = perms.filter_by(division=division)
        return db.session.query(perms.exists()).all()[0][0]


class APIKey(db.Model, AutoID, AutoName, Timestamped):

    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    user = db.relationship('User', back_populates='api_keys')

    key = db.Column(db.LargeBinary(32), nullable=False)

    def __init__(self, user):
        self.user = user
        self.key = os.urandom(32)

    @property
    def hex_key(self):
        return urlsafe_b64encode(self.key).decode('utf-8').replace(u'=', u',')


class User(Entity):
    """User base class.

    Represents users who can submit, review and/or pay out requests. It also
    supplies a number of convenience methods for subclasses.
    """
    id = db.Column(db.Integer, db.ForeignKey('entity.id'), primary_key=True)

    #: If the user is an administrator. This allows the user to create and
    #: administer divisions.
    admin = db.Column(db.Boolean, nullable=False, default=False)

    #: :py:class:`~.Request`\s this user has submitted.
    requests = db.relationship(Request, back_populates='submitter')

    #: :py:class:`~.Action`\s this user has performed on requests.
    actions = db.relationship(Action, back_populates='user')

    #: :py:class:`~.Pilot`\s associated with this user.
    pilots = db.relationship('Pilot', back_populates='user',
            collection_class=set)

    #: :py:class:`Group`\s this user is a member of
    groups = db.relationship('Group', secondary=users_groups,
            back_populates='users', collection_class=set)

    notes = db.relationship('Note', back_populates='user',
            order_by='desc(Note.timestamp)', foreign_keys='Note.user_id')

    notes_made = db.relationship('Note', back_populates='noter',
            order_by='desc(Note.timestamp)', foreign_keys='Note.noter_id')

    api_keys = db.relationship(APIKey, back_populates='user')

    @hybrid_property
    def permissions(self):
        """All :py:class:`Permission` objects associated with this user."""
        groups = db.session.query(users_groups.c.group_id.label('group_id'))\
                .filter(users_groups.c.user_id==self.id).subquery()
        group_perms = db.session.query(Permission)\
                .join(groups, groups.c.group_id==Permission.entity_id)
        user_perms = db.session.query(Permission)\
                .join(User)\
                .filter(User.id==self.id)
        perms = user_perms.union(group_perms)
        return perms

    @permissions.expression
    def permissions(cls):
        groups = db.select([users_groups.c.group_id])\
                .where(users_groups.c.user_id==cls.id).alias()
        group_permissions = db.select([Permission])\
                .where(Permission.entity_id.in_(groups)).alias()
        user_permissions = db.select([Permission])\
                .where(Permission.entity_id==cls.id)
        return user_permissions.union(group_permissions)

    def is_authenticated(self):
        """Part of the interface for Flask-Login."""
        return True

    def is_active(self):
        """Part of the interface for Flask-Login."""
        return True

    def is_anonymous(self):
        """Part of the interface for Flask-Login."""
        return False

    def get_id(self):
        """Part of the interface for Flask-Login."""
        return str(self.id)

    def submit_divisions(self):
        """Get a list of the divisions this user is able to submit requests to.

        :returns: A list of tuples. The tuples are in the form (division.id,
            division.name)
        :rtype: list
        """
        submit_perms = self.permissions\
                .filter_by(permission=PermissionType.submit)\
                .subquery()
        divisions = db.session.query(Division).join(submit_perms)\
                .order_by(Division.name)
        # Remove duplicates and sort divisions by name
        choices = []
        for name, group in groupby(divisions, lambda d: d.name):
            choices.append((six.next(group).id, name))
        return choices


class Note(db.Model, AutoID, Timestamped, AutoName):

    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    user = db.relationship(User, back_populates='notes',
            foreign_keys=[user_id])

    content = db.Column(db.Text(convert_unicode=True), nullable=False)

    noter_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    noter = db.relationship(User, back_populates='notes_made',
            foreign_keys=[noter_id])

    def __init__(self, user, noter, note):
        self.user = user
        self.noter = noter
        self.content = ensure_unicode(note)


@unistr
class Pilot(db.Model, AutoID, AutoName):
    """Represents an in-game character."""

    # Character names in Eve are resticted to ASCII, but use unicode for
    # consistency with the rest of the database columns (and if they ever
    # decide to lift this restriction).
    #: The name of the character
    name = db.Column(db.String(150, convert_unicode=True), nullable=False)

    #: The id of the User this character belongs to.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The User this character belongs to.
    user = db.relationship(User, back_populates='pilots')

    #: The Requests filed with lossmails from this character.
    requests = db.relationship(Request, back_populates='pilot',
            collection_class=list, order_by=Request.timestamp.desc())

    def __init__(self, user, name, id_):
        """Create a new Pilot instance.

        :param user: The user this character belpongs to.
        :type user: :py:class:`~.User`
        :param str name: The name of this character.
        :param int id_: The CCP-given characterID number.
        """
        self.user = user
        self.name = ensure_unicode(name)
        self.id = id_

    def __repr__(self):
        return "{x.__class__.__name__}({x.user}, '{x.name}', {x.id})".format(
                x=self)

    def __unicode__(self):
        return self.name


class Group(Entity):
    """Base class for a group of users.

    Represents a group of users. Usable for granting permissions to submit,
    evaluate and pay.
    """

    id = db.Column(db.Integer, db.ForeignKey('entity.id'), primary_key=True)

    #: :py:class:`User` s that belong to this group.
    users = db.relationship(User, secondary=users_groups,
            back_populates='groups', collection_class=set)

    #: Synonym for :py:attr:`entity_permissions`
    permissions = db.synonym('entity_permissions')


class Permission(db.Model, AutoID, AutoName):
    __table_args__ = (
        db.UniqueConstraint('division_id', 'entity_id', 'permission'),
    )

    division_id = db.Column(db.Integer, db.ForeignKey('division.id'),
            nullable=False)

    #: The division this permission is granting access to
    division = db.relationship('Division',
            back_populates='division_permissions')

    entity_id = db.Column(db.Integer, db.ForeignKey('entity.id'),
            nullable=False)

    #: The :py:class:`Entity` being granted access
    entity = db.relationship(Entity, back_populates='entity_permissions')

    #: The permission being granted.
    permission = db.Column(PermissionType.db_type(), nullable=False)

    def __init__(self, division, permission, entity):
        """Create a Permission object granting an entity access to a division.
        """
        self.division = division
        self.entity = entity
        self.permission = permission

    def __repr__(self):
        return ("{x.__class__.__name__}('{x.permission}', {x.entity}, "
               "{x.division})").format(x=self)


class TransformerRef(db.Model, AutoID, AutoName):
    """Stores associations between :py:class:`~.Transformer`\s and
    :py:class:`.Division`\s.
    """

    __table_args__ = (
        db.UniqueConstraint('division_id', 'attribute_name'),
    )

    #: The attribute this transformer is applied to.
    attribute_name = db.Column(db.String(50), nullable=False)

    #: The transformer instance.
    transformer = db.Column(db.PickleType, nullable=True)

    division_id = db.Column(db.Integer, db.ForeignKey('division.id'),
            nullable=False)

    #: The division the transformer is associated with
    division = db.relationship('Division',
            back_populates='division_transformers')

    @db.validates('transformer')
    def prune_null_transformers(self, attr, transformer):
        """Removes :py:class:`TransformerRef`\s when :py:attr:`.transformer` is
        removed.
        """
        if transformer is None:
            self.division = None
        return transformer


class Division(db.Model, AutoID, AutoName):
    """A reimbursement division.

    A division has (possibly non-intersecting) groups of people that can submit
    requests, review requests, and pay out requests.
    """

    #: The name of this division.
    name = db.Column(db.String(128, convert_unicode=True), nullable=False)

    #: All :py:class:`Permission`\s associated with this division.
    division_permissions = db.relationship(Permission, collection_class=set,
            back_populates='division')

    #: :py:class:`Request` s filed under this division.
    requests = db.relationship(Request, back_populates='division')

    division_transformers = db.relationship(TransformerRef,
            collection_class=attribute_mapped_collection('attribute_name'),
            back_populates='division', cascade='delete,delete-orphan')

    #: A mapping of attribute names to :py:class:`~.transformer.Transformer`
    #: instances.
    transformers = association_proxy(
            'division_transformers',
            'transformer',
            creator=lambda attr, trans:
                    TransformerRef(attribute_name=attr, transformer=trans))

    @property
    def permissions(self):
        """The permissions objects for this division, mapped via their
        permission names.
        """
        class _PermProxy(object):
            def __init__(self, perms):
                self.perms = perms
            def __getitem__(self, key):
                return set(filter(lambda x: x.permission == key, self.perms))
        return _PermProxy(self.division_permissions)

    def __init__(self, name):
        self.name = ensure_unicode(name)

    def __repr__(self):
        return "{x.__class__.__name__}('{x.name}')".format(x=self)



"""Move from using floats for ISK to numeric types.

Revision ID: 4198a248c8a
Revises: 45024170cf6
Create Date: 2014-06-18 14:34:25.967159

"""

# revision identifiers, used by Alembic.
revision = '4198a248c8a'
down_revision = '45024170cf6'

from decimal import Decimal
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column


def upgrade():
    op.add_column('request',
            sa.Column('numeric_base_payout', sa.Numeric(precision=15, scale=2),
                    default=0.0)
    )
    request = table('request',
            column('id', sa.Integer),
            column('base_payout', sa.Float),
            column('numeric_base_payout', sa.Numeric(precision=15, scale=2)),
    )
    conn = op.get_bind()
    requests_sel = select([request.c.id, request.c.base_payout])
    requests = conn.execute(requests_sel)
    for request_id, float_payout in requests:
        decimal_payout = Decimal.from_float(float_payout)
        decimal_payout *= 1000000
        update_stmt = update(request)\
                .where(request.c.id == request_id)\
                .values({
                    'numeric_base_payout': decimal_payout,
                })
        conn.execute(update_stmt)
    requests.close()
    op.drop_column('request', 'base_payout')
    op.alter_column('request',
            column_name='numeric_base_payout',
            new_column_name='base_payout',
            existing_type=sa.Numeric(precision=15, scale=2),
            existing_server_default=0.0)


def downgrade():
    op.add_column('request',
            sa.Column('float_base_payout', sa.Float, default=0.0)
    )
    request = table('request',
            column('id', sa.Integer),
            column('base_payout', sa.Numeric(precision=15, scale=2)),
            column('float_base_payout', sa.Float),
    )
    conn = op.get_bind()
    requests_sel = select([request.c.id, request.c.base_payout])
    requests = conn.execute(requests_sel)
    for request_id, decimal_payout in requests:
        decimal_payout = decimal_payout / 1000000
        float_payout = float(decimal_payout)
        update_stmt = update(request)\
                .where(request.c.id == request_id)\
                .values({
                    'float_base_payout': float_payout,
                })
        conn.execute(update_stmt)
    requests.close()
    op.drop_column('request', 'base_payout')
    op.alter_column('request',
            column_name='numeric_base_payout',
            new_column_name='base_payout',
            existing_type=sa.Float,
            existing_server_default=0.0)



from __future__ import absolute_import
from collections import OrderedDict
import re

from flask import render_template, abort, url_for, flash, Markup, request,\
    redirect, current_app, Blueprint, Markup, json, make_response
from flask.views import View
from flask.ext.login import login_required, current_user
from flask.ext.wtf import Form
import six
from six.moves import map
from wtforms.fields import SelectField, SubmitField, TextAreaField, HiddenField
from wtforms.fields.html5 import URLField, DecimalField
from wtforms.validators import InputRequired, AnyOf, URL, ValidationError,\
    StopValidation

from .. import db
from ..models import Request, Modifier, Action, ActionType, ActionError,\
        ModifierError, AbsoluteModifier, RelativeModifier
from ..util import xmlify, jsonify
from ..auth import PermissionType
from ..auth.models import Division, Pilot, Permission, User, Group, Note,\
    APIKey


if six.PY3:
    unicode = str


blueprint = Blueprint('requests', __name__)


class RequestListing(View):
    """Abstract class for lists of :py:class:`~evesrp.models.Request`\s.

    Subclasses will be able to respond to both normal HTML requests as well as
    to API requests with JSON.
    """

    #: The template to use for listing requests
    template = 'list_requests.html'

    #: Decorators to apply to the view functions
    decorators = [login_required]

    def requests(self, division_id=None):
        """Returns a list :py:class:`~.Request`\s belonging to
        the specified :py:class:`~.Division`, or all divisions if
        ``None``. Must be implemented by subclasses, as this is an abstract
        method.

        :param int division_id: ID number of a :py:class:`~.Division`, or
            ``None``.
        :returns: :py:class:`~.models.Request`\s
        :rtype: iterable
        """
        raise NotImplementedError()

    def dispatch_request(self, division_id=None, page=1, **kwargs):
        """Returns the response to requests.

        Part of the :py:class:`flask.views.View` interface.
        """
        if request.is_json or request.is_xhr:
            return jsonify(requests=self.requests(division_id))
        if request.is_rss:
            return xmlify('rss.xml', content_type='application/rss+xml',
                    requests=self.requests(division_id),
                    title=(kwargs['title'] if 'title' in kwargs else u''),
                    main_link=url_for(request.endpoint,
                        division_id=division_id, _external=True))
        if request.is_xml:
            return xmlify('request_list.xml',
                    requests=self.requests(division_id))
        pager = self.requests(division_id).paginate(page, per_page=20)
        return render_template(self.template,
                pager=pager, **kwargs)

    @property
    def _load_options(self):
        """Returns a sequence of
        :py:class:`~sqlalchemy.orm.strategy_options.Load` objects specifying
        which attributes to load (or really any load options necessary).
        """
        return (
                db.Load(Request).load_only('id', 'pilot_id', 'division_id',
                    'system', 'ship_type', 'status', 'timestamp',
                    'base_payout'),
                db.Load(Division).joinedload('name'),
                db.Load(Pilot).joinedload('name'),
        )


class APIKeyForm(Form):
    action = HiddenField(validators=[AnyOf(['add', 'delete'])])
    key_id = HiddenField()


class PersonalRequests(RequestListing):
    """Shows a list of all personally submitted requests and divisions the user
    has permissions in.

    It will show all requests the current user has submitted.
    """

    template = 'personal.html'

    methods = ['GET', 'POST']

    def dispatch_request(self, division_id=None, page=1, **kwargs):
        if request.method == 'POST':
            form = APIKeyForm()
            if form.validate():
                if form.action.data == 'add':
                    key = APIKey(current_user)
                else:
                    key = APIKey.query.get(int(form.key_id.data))
                    if key is not None:
                        db.session.delete(key)
                db.session.commit()
        # Handle API access in here, so we can add extra data
        if request.is_json or request.is_xhr:
            return jsonify(
                    requests=self.requests(division_id),
                    api_keys=current_user.api_keys)
        if request.is_xml:
            return xmlify('personal_list.xml',
                    requests=self.requests(division_id))
        return super(PersonalRequests, self).dispatch_request(
                division_id, page, key_form=APIKeyForm(formdata=None))

    def requests(self, division_id=None):
        requests = Request.query\
                .join(User)\
                .filter(User.id==current_user.id)\
                .options(*self._load_options)
        if division_id is not None:
            requests = requests.filter(Request.division_id==division_id)
        requests = requests.order_by(Request.timestamp.desc())
        return requests


class PermissionRequestListing(RequestListing):
    """Show all requests that the current user has permissions to access.

    This is used for the various permission-specific views.
    """

    def __init__(self, permissions, statuses):
        """Create a :py:class:`PermissionRequestListing` for the given
        permissions and statuses.

        :param tuple permissions: The permissions to filter by
        :param tuple statuses: A tuple of valid statuses for requests to be in
        """
        if permissions in PermissionType.all:
            permissions = (permissions,)
        # Admin permission has to be explicitly added because it's used in a
        # complicated query in requests()
        self.permissions = (PermissionType.admin,) + tuple(permissions)
        self.statuses = statuses

    def dispatch_request(self, division_id=None, page=1, **kwargs):
        if not current_user.has_permission(self.permissions):
            abort(403)
        else:
            if 'title' in kwargs:
                title = kwargs.pop('title')
            else:
                title = u', '.join(map(lambda s: s.description, self.statuses))
            return super(PermissionRequestListing, self).dispatch_request(
                    division_id,
                    page,
                    title=title,
                    **kwargs)

    def requests(self, division_id=None):
        user_perms = db.session.query(Permission.id.label('permission_id'),
                Permission.division_id.label('division_id'),
                Permission.permission.label('permission'))\
                .filter(Permission.entity==current_user)
        group_perms = db.session.query(Permission.id.label('permission_id'),
                Permission.division_id.label('division_id'),
                Permission.permission.label('permission'))\
                .join(Group)\
                .filter(Group.users.contains(current_user))
        perms = user_perms.union(group_perms)\
                .filter(Permission.permission.in_(self.permissions))
        if division_id is not None:
            perms = perms.filter(Permission.division_id==division_id)
        perms = perms.subquery()
        requests = Request.query\
                .join(perms, Request.division_id==perms.c.division_id)\
                .filter(Request.status.in_(self.statuses))\
                .order_by(Request.timestamp.desc())\
                .options(*self._load_options)
        return requests


class PayoutListing(PermissionRequestListing):
    """A special view made for quickly processing payouts for requests."""

    template = 'payout.html'

    def __init__(self):
        # Just a special case of PermissionRequestListing
        super(PayoutListing, self).__init__((PermissionType.pay,),
                (ActionType.approved,))

    def dispatch_request(self, division_id=None, page=1):
        if not current_user.has_permission(self.permissions):
            abort(403)
        return super(PayoutListing, self).dispatch_request(
                division_id,
                page,
                title=u', '.join(map(lambda s: s.description, self.statuses)),
                form=ActionForm())


def register_perm_request_listing(app, endpoint, path, permissions, statuses):
    """Utility function for creating :py:class:`PermissionRequestListing`
    views.

    :param app: The application to add the view to
    :type app: :py:class:`flask.Flask`
    :param str endpoint: The name of the view
    :param str path: The URL path for the view
    :param tuple permissions: Passed to
        :py:meth:`PermissionRequestListing.__init__`
    :param iterable statuses: Passed to
        :py:meth:`PermissionRequestListing.__init__`
    """
    if not path.endswith('/'):
        path += '/'
    view = PermissionRequestListing.as_view(endpoint, permissions=permissions,
            statuses=statuses)
    app.add_url_rule(path, view_func=view)
    app.add_url_rule('{}rss.xml'.format(path), view_func=view)
    if path != '/':
        app.add_url_rule('{}<int:division_id>/rss.xml'.format(path),
                view_func=view)
        app.add_url_rule('{}<int:page>/'.format(path), view_func=view)
        app.add_url_rule('{}<int:page>/<int:division_id>/'.format(path),
                view_func=view)


@blueprint.record
def register_class_views(state):
    """Called when the blueprint is registered, this function defines routes
    for and attaches the class-based views to the app.
    """
    try:
        prefixes = state.app.request_prefixes
    except AttributeError:
        prefixes = []
        state.app.request_prefixes = prefixes
    prefixes.append(state.url_prefix if state.url_prefix is not None else '')
    # Personal list
    personal_view = PersonalRequests.as_view('personal_requests')
    state.add_url_rule('/personal/', view_func=personal_view)
    state.add_url_rule('/personal/rss.xml', view_func=personal_view)
    state.add_url_rule('/personal/<int:division_id>/rss.xml',
            view_func=personal_view)
    state.add_url_rule('/personal/<int:page>/', view_func=personal_view)
    state.add_url_rule('/personal/<int:page>/<int:division_id>',
            view_func=personal_view)
    # Payout list
    payout_view = PayoutListing.as_view('list_approved_requests')
    payout_url_stub = '/pay/'
    state.add_url_rule(payout_url_stub, view_func=payout_view)
    state.add_url_rule(payout_url_stub + 'rss.xml', view_func=payout_view)
    state.add_url_rule(payout_url_stub + '<int:division_id>/rss.xml',
            view_func=payout_view)
    state.add_url_rule(payout_url_stub + '<int:page>/', view_func=payout_view)
    state.add_url_rule(payout_url_stub + '<int:page>/<int:division_id>/',
            view_func=payout_view)
    # Other more generalized listings
    register_perm_request_listing(state, 'list_pending_requests',
            '/pending/', (PermissionType.review,), ActionType.pending)
    register_perm_request_listing(state, 'list_completed_requests',
            '/completed/', PermissionType.elevated, ActionType.finalized)
    # Special all listing, mainly intended for API users
    register_perm_request_listing(state, 'list_all_requests',
            '/', PermissionType.elevated, ActionType.statuses)


class ValidKillmail(URL):
    """Custom :py:class:'~.Field' validator that checks if any
    :py:class:`~.Killmail` accepts the given URL.
    """
    def __init__(self, mail_class, **kwargs):
        self.mail_class = mail_class
        super(ValidKillmail, self).__init__(**kwargs)

    def __call__(self, form, field):
        super(ValidKillmail, self).__call__(form, field)
        try:
            mail = self.mail_class(field.data)
        except ValueError as e:
            if six.PY2:
                raise ValidationError(unicode(e))
        except LookupError as e:
            if six.PY2:
                raise ValidationError(unicode(e))
        else:
            if mail.verified:
                form.killmail = mail
                raise StopValidation
            else:
                raise ValidationError(
                        u'{} cannot be verified.'.format(field.data))


def get_killmail_validators():
    """Get a list of :py:class:`ValidKillmail`\s for each killmail source.

    This method is used to delay accessing `current_app` until we're in a
    request context.
    :returns: a list of :py:class:`ValidKillmail`\s
    :rtype list:
    """
    validators = [ValidKillmail(s) for s in current_app.killmail_sources]
    validators.append(InputRequired())
    return validators


def get_killmail_descriptions():
    description = Markup(u"Acceptable Killmail Links:<ul>")
    desc_entry = Markup(u"<li>{}</li>")
    killmail_descs = [desc_entry.format(km.description) for km in\
            current_app.killmail_sources]
    description += Markup(u"").join(killmail_descs)
    description += Markup(u"</ul>")
    return description


class RequestForm(Form):
    url = URLField(u'Killmail URL')
    details = TextAreaField(u'Details', validators=[InputRequired()],
            description=u'Supporting details about your loss.')
    division = SelectField(u'Division', coerce=int)
    submit = SubmitField(u'Submit')

    def validate_url(form, field):
        failures = set()
        for v in get_killmail_validators():
            try:
                v(form, field)
            except ValidationError as e:
                failures.add(unicode(e))
            else:
                continue
        else:
            # If execution reached here, it means a StopValidation exception
            # wasn't raised (meaning the killmail isn't valid).
            raise ValidationError([e for e in failures])


@blueprint.route('/add/', methods=['GET', 'POST'])
@login_required
def submit_request():
    """Submit a :py:class:`~.models.Request`\.

    Displays a form for submitting a request and then processes the submitted
    information. Verifies that the user has the appropriate permissions to
    submit a request for the chosen division and that the killmail URL given is
    valid. Also enforces that the user submitting this requests controls the
    character from the killmail and prevents duplicate requests.
    """
    if not current_user.has_permission(PermissionType.submit):
        abort(403)
    form = RequestForm()
    # Do it in here so we can access current_app (needs to be in an app
    # context)
    form.url.description = get_killmail_descriptions()
    # Create a list of divisions this user can submit to
    form.division.choices = current_user.submit_divisions()

    if form.validate_on_submit():
        mail = form.killmail
        # Prevent submitting other people's killmails
        pilot = Pilot.query.get(mail.pilot_id)
        if not pilot or pilot not in current_user.pilots:
            flash(u"You can only submit killmails of characters you control",
                    u'warning')
            return render_template('form.html', form=form)
        # Prevent duplicate killmails
        # The name 'request' is already used by Flask.
        # Hooray name collisions!
        srp_request = Request.query.get(mail.kill_id)
        if srp_request is None:
            division = Division.query.get(form.division.data)
            srp_request = Request(current_user, form.details.data, division,
                    mail)
            srp_request.pilot = pilot
            db.session.add(srp_request)
            db.session.commit()
            return redirect(url_for('.get_request_details',
                request_id=srp_request.id))
        else:
            flash(u"This kill has already been submitted", u'warning')
            return redirect(url_for('.get_request_details',
                request_id=srp_request.id))
    return render_template('form.html', form=form)


class ModifierForm(Form):
    id_ = HiddenField(default='modifier')
    value = DecimalField(u'Value')
    # TODO: add a validator for the type
    type_ = HiddenField(validators=[AnyOf(('rel-bonus', 'rel-deduct',
            'abs-bonus', 'abs-deduct'))])
    note = TextAreaField(u'Reason')


class VoidModifierForm(Form):
    id_ = HiddenField(default='void')
    modifier_id = HiddenField()
    void = SubmitField(Markup(u'x'))

    def __init__(self, modifier=None, *args, **kwargs):
        if modifier is not None:
            self.modifier_id = modifier.id
        super(VoidModifierForm, self).__init__(*args, **kwargs)


class PayoutForm(Form):
    id_ = HiddenField(default='payout')
    value = DecimalField(u'M ISK', validators=[InputRequired()])


class ActionForm(Form):
    id_ = HiddenField(default='action')
    note = TextAreaField(u'Note')
    type_ = HiddenField(default='comment',
            validators=[AnyOf(list(ActionType.values()))])



class ChangeDetailsForm(Form):
    id_ = HiddenField(default='details')
    details = TextAreaField(u'Details', validators=[InputRequired()])


class AddNote(Form):
    id_ = HiddenField(default='note')
    note = TextAreaField(u'Add Note',
            description=( "If you have something like '#{Kill ID}', it will be"
                         u" linkified to the corresponding request "
                         u"(if it exists). For example, #1234567 would be "
                         u"linked to the request for the kill with ID "
                         u"1234567."),
            validators=[InputRequired()])


killmail_re = re.compile(r'#(\d+)')


@blueprint.route('/<int:request_id>/', methods=['GET'])
@login_required
def get_request_details(request_id=None, srp_request=None):
    """Handles responding to all of the :py:class:`~.models.Request` detail
    functions.

    The various modifier functions all depend on this function to create the
    actual response content.
    Only one of the arguments is required. The ``srp_request`` argument is a
    conveniece to other functions calling this function that have already
    retrieved the request.

    :param int request_id: the ID of the request.
    :param srp_request: the request.
    :type srp_request: :py:class:`~.models.Request`
    """
    if srp_request is None:
        srp_request = Request.query.get_or_404(request_id)
    # Different templates are used for different roles
    if current_user.has_permission(PermissionType.review,
            srp_request.division):
        template = 'request_review.html'
    elif current_user.has_permission(PermissionType.pay, srp_request.division):
        template = 'request_pay.html'
    elif current_user == srp_request.submitter:
        template = 'request_detail.html'
    else:
        abort(403)
    if request.is_json:
        # dump the load to encode srp_request as json and then get a dictionary
        # form of it. We need this to add a few bits of information to the
        # standard request encoding
        enc_request = json.loads(json.dumps(srp_request))
        enc_request[u'actions'] = srp_request.actions
        enc_request[u'modifiers'] = srp_request.modifiers
        valid_actions = map(
                lambda a: a.value,
                srp_request.valid_actions(current_user))
        enc_request[u'valid_actions'] = valid_actions
        enc_request[u'current_user'] = current_user._get_current_object()
        return jsonify(enc_request)
    if request.is_xml:
        return xmlify('request.xml', srp_request=srp_request)
    return render_template(template, srp_request=srp_request,
            modifier_form=ModifierForm(formdata=None),
            payout_form=PayoutForm(formdata=None),
            action_form=ActionForm(formdata=None),
            void_form=VoidModifierForm(formdata=None),
            details_form=ChangeDetailsForm(formdata=None, obj=srp_request),
            note_form=AddNote(formdata=None))


def _add_modifier(srp_request):
    form = ModifierForm()
    if form.validate():
        if 'bonus' in form.type_.data:
            value = form.value.data
        elif 'deduct' in form.type_.data:
            value = form.value.data * -1
        if 'abs' in form.type_.data:
            ModClass = AbsoluteModifier
            value *= 1000000
        elif 'rel' in form.type_.data:
            ModClass = RelativeModifier
            value /= 100
        try:
            mod = ModClass(srp_request, current_user, form.note.data, value)
            db.session.add(mod)
            db.session.commit()
        except ModifierError as e:
            flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _change_payout(srp_request):
    form = PayoutForm()
    if not current_user.has_permission(PermissionType.review, srp_request):
        flash(u"Only reviewers can change the base payout.", u'error')
    elif form.validate():
        try:
            srp_request.base_payout = form.value.data * 1000000
            db.session.commit()
        except ModifierError as e:
            flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _add_action(srp_request):
    form = ActionForm()
    if form.validate():
        type_ = ActionType.from_string(form.type_.data)
        try:
            Action(srp_request, current_user, form.note.data, type_)
            db.session.commit()
        except ActionError as e:
            flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _void_modifier(srp_request):
    form = VoidModifierForm()
    if form.validate():
        modifier_id = int(form.modifier_id.data)
        modifier = Modifier.query.get(modifier_id)
        if modifier is None:
            flash(u"Invalid modifier ID {}.".format(modifier_id),
                    u'error')
        else:
            try:
                modifier.void(current_user)
                db.session.commit()
            except ModifierError as e:
                flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _change_details(srp_request):
    form = ChangeDetailsForm()
    if current_user != srp_request.submitter:
        flash(u"Only the submitter can change the request details.", u'error')
    elif srp_request.finalized:
        flash(u"Details con only be changed when the request is still "
                u"pending.", u'error')
    elif form.validate():
        archive_note = u"Old Details: " + srp_request.details
        archive_action = Action(srp_request, current_user, archive_note)
        archive_action.type_ = ActionType.evaluating
        srp_request.details = form.details.data
        db.session.commit()
    return get_request_details(srp_request=srp_request)


def _add_note(srp_request):
    form = AddNote()
    if not current_user.has_permission(PermissionType.elevated):
        flash(u"You do not have permission to add a note to a user.", u'error')
    elif form.validate():
        # Linkify killmail IDs
        note_content = Markup.escape(form.note.data)
        for match in killmail_re.findall(note_content):
            kill_id = int(match)
            check_request = db.session.query(Request.id).filter_by(id=kill_id)
            if db.session.query(check_request.exists()):
                link = u'<a href="{url}">#{kill_id}</a>'.format(
                        url=url_for('.get_request_details',
                                request_id=kill_id),
                        kill_id=kill_id)
                link = Markup(link)
                note_content = note_content.replace(u'#' + match, link)
        # Create the note
        note = Note(srp_request.submitter, current_user, note_content)
        db.session.commit()
    return get_request_details(srp_request=srp_request)


@blueprint.route('/<int:request_id>/', methods=['POST'])
@login_required
def modify_request(request_id):
    """Handles POST requests that modify :py:class:`~.models.Request`\s.

    Because of the numerous possible forms, this function bounces execution to
    a more specific function base on the form's "id_" field.

    :param int request_id: the ID of the request.
    """
    srp_request = Request.query.get_or_404(request_id)
    if request.form['id_'] == 'modifier':
        return _add_modifier(srp_request)
    elif request.form['id_'] == 'payout':
        return _change_payout(srp_request)
    elif request.form['id_'] == 'action':
        return _add_action(srp_request)
    elif request.form['id_'] == 'void':
        return _void_modifier(srp_request)
    elif request.form['id_'] == 'details':
        return _change_details(srp_request)
    elif request.form['id_'] == 'note':
        return _add_note(srp_request)
    else:
        return abort(400)


class DivisionChange(Form):
    division = SelectField(u'Divisions', coerce=int)
    submit = SubmitField(u'Submit')


@blueprint.route('/<int:request_id>/division/', methods=['GET', 'POST'])
@login_required
def request_change_division(request_id):
    srp_request = Request.query.get_or_404(request_id)
    if not current_user.has_permission(PermissionType.review, srp_request) and\
            current_user != srp_request.submitter:
        current_app.logger.warn(u"User '{}' does not have permission to change"
                                u" request #{}'s division".format(
                                    current_user, srp_request.id))
        abort(403)
    if srp_request.finalized:
        msg = (u"Cannot change request #{}'s division as it is in a finalized"
               u" state").format(srp_request.id)
        current_app.logger.info(msg)
        flash(u"Cannot change the division as this request is in a finalized"
              u" state", u'warning')
        return redirect(url_for('.get_request_details', request_id=request_id))
    division_choices = srp_request.submitter.submit_divisions()
    try:
        division_choices.remove(
                (srp_request.division.id, srp_request.division.name))
    except ValueError:
        pass
    if len(division_choices) == 0:
        current_app.logger.debug(u"No other divisions to move request #{} to."\
                .format(srp_request.id))
        flash(u"No other divisions to move to.", u'info')
        return redirect(url_for('.get_request_details', request_id=request_id))
    form = DivisionChange()
    form.division.choices = division_choices
    if form.validate_on_submit():
        new_division = Division.query.get(form.division.data)
        archive_note = u"Moving from division '{}' to division '{}'.".format(
                srp_request.division.name,
                new_division.name)
        archive_action = Action(srp_request, current_user, archive_note)
        archive_action.type_ = ActionType.evaluating
        srp_request.division = new_division
        db.session.commit()
        flash(u'Request #{} moved to {} division'.format(srp_request.id,
                new_division.name), u'success')
        if current_user.has_permission(PermissionType.elevated, new_division) \
                or current_user == srp_request.submitter:
            return redirect(url_for('.get_request_details',
                    request_id=request_id))
        else:
            return redirect(url_for('.list_pending_requests'))
    else:
        current_app.logger.warn(u"Form validation failed for division change:"
                                u" {}.".format(form.errors))
    form.division.data = srp_request.division.id
    return render_template('form.html', form=form)



from __future__ import absolute_import
from base64 import urlsafe_b64decode
import binascii
from flask import render_template, url_for, abort, session, redirect, request,\
        current_app, g, Blueprint
from flask.ext.login import login_required, logout_user, LoginManager
from six.moves import map
from sqlalchemy.orm.exc import NoResultFound
from .. import csrf
from ..auth.models import User, APIKey
from ..util import ensure_unicode


blueprint = Blueprint('login', __name__)


login_manager = LoginManager()


@login_manager.user_loader
def login_loader(userid):
    """Pull a user object from the database.

    This is used for loading users from existing sessions.
    """
    return User.query.get(int(userid))


@login_manager.request_loader
def apikey_loader(request):
    api_key = ensure_unicode(request.values.get('apikey'))
    if api_key and request.method == 'GET':
        api_key = api_key.replace(u',', u'=')
        try:
            api_key = urlsafe_b64decode(api_key.encode('utf-8'))
        except binascii.Error:
            # If the api key is malformed, binascii throws an exception.
            # Rejected.
            return None
        try:
            key = APIKey.query.filter_by(key=api_key).one()
        except NoResultFound:
            pass
        else:
            return key.user

    # returning None signifies failure for this method
    return None


@blueprint.route('/login/', methods=['GET', 'POST'])
def login():
    """Presents the login form and processes responses from that form.

    When a POST request is recieved, this function passes control to the
    appropriate :py:meth:`login <evesrp.auth.AuthMethod.login>` method.
    """
    # forms is a list of tuples. The tuples are
    # (AuthMethod instance, AuthForm instance)
    forms = []
    for auth_method in current_app.auth_methods:
        prefix = auth_method.safe_name
        form = auth_method.form()
        forms.append((auth_method, form(prefix=prefix)))
    if request.method == 'POST':
        # Find the form that was submitted. The unique prefix for each form
        # means only one form.submit is going to be valid.
        for auth_tuple in forms:
            if auth_tuple[1].submit.data:
                auth_method, form = auth_tuple
                break
        else:
            abort(400)
        if form.validate():
            return auth_method.login(form)
    return render_template('login.html', forms=forms)


login_manager.login_view = 'login.login'


# 302 redirects let the request method change to GET if it started as POST.
# By defining routes for both of these paths, the implicit redirect for the
# first route is skipped.
@blueprint.route('/login/<string:auth_method>', methods=['GET', 'POST'])
@blueprint.route('/login/<string:auth_method>/', methods=['GET', 'POST'])
def auth_method_login(auth_method):
    """Trampoline for :py:class:`~evesrp.auth.AuthMethod`\-specific views.

    See :py:meth:`Authmethod.view <evesrp.auth.AuthMethod.view>` for more
    details.
    """
    method_map = dict(map((lambda m: (m.safe_name, m)),
        current_app.auth_methods))
    return method_map[auth_method].view()


@blueprint.route('/logout/')
@login_required
def logout():
    """Logs the current user out.

    Redirects to :py:func:`.index`.
    """
    logout_user()
    return redirect(url_for('index'))



"""Generalize attribute transformers for divisions.

Revision ID: 4280bf2417c
Revises: 2976d59f286
Create Date: 2014-06-18 22:01:20.924226

"""

# revision identifiers, used by Alembic.
revision = '4280bf2417c'
down_revision = '2976d59f286'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column, or_
import evesrp.transformers


division = table('division',
        column('id', sa.Integer),
        column('ship_transformer', sa.PickleType),
        column('pilot_transformer', sa.PickleType),
)

transformerref = table('transformerref',
        column('division_id', sa.Integer),
        column('attribute_name', sa.String(length=50)),
        column('transformer', sa.PickleType),
)


# This is tricky: Ensure that evesrp.transformers has ShipTransformer and
# PilotTransformer classes so pickle con unpack them
for legacy_transformer in ('ShipTransformer', 'PilotTransformer'):
    if not hasattr(evesrp.transformers, legacy_transformer):
        new_class = type(legacy_transformer,
                (evesrp.transformers.Transformer,), {})
        setattr(evesrp.transformers, legacy_transformer, new_class)


def upgrade():
    # Create new transformerref table
    op.create_table('transformerref',
            sa.Column('id', sa.Integer, nullable=False, primary_key=True),
            sa.Column('attribute_name', sa.String(length=50), nullable=False),
            sa.Column('transformer', sa.PickleType, nullable=False),
            sa.Column('division_id', sa.Integer, nullable=False),
            sa.ForeignKeyConstraint(['division_id'], ['division.id'], ),
            sa.UniqueConstraint('division_id', 'attribute_name')
    )
    # Migrate ship and pilot transformers
    conn = op.get_bind()
    columns = [division.c.id, division.c.ship_transformer,
        division.c.pilot_transformer]
    transformer_sel = select(columns)\
            .where(or_(
                    division.c.ship_transformer != None,
                    division.c.pilot_transformer != None
            ))
    transformer_rows = conn.execute(transformer_sel)
    new_transformers = []
    for division_id, ship_transformer, pilot_transformer in transformer_rows:
        if ship_transformer is not None:
            transformer = evesrp.transformers.Transformer(
                    ship_transformer.name,
                    ship_transformer.slug)
            new_transformers.append({
                    'attribute_name': 'ship_type',
                    'transformer': transformer,
                    'division_id': division_id,
            })
        if pilot_transformer is not None:
            transformer = evesrp.transformers.Transformer(
                    pilot_transformer.name,
                    pilot_transformer.slug)
            new_transformers.append({
                    'attribute_name': 'pilot',
                    'transformer': transformer,
                    'division_id': division_id,
            })
    transformer_rows.close()
    op.bulk_insert(transformerref, new_transformers)
    # Drop old columns
    op.drop_column('division', 'ship_transformer')
    op.drop_column('division', 'pilot_transformer')


def downgrade():
    # Add ship and pilot transformer columns back to division
    op.add_column('division', sa.Column('ship_transformer', sa.PickleType))
    op.add_column('division', sa.Column('pilot_transformer', sa.PickleType))
    # Convert transformerrefs back to the old columns
    conn = op.get_bind()
    columns = [
        transformerref.c.division_id,
        transformerref.c.attribute_name,
        transformerref.c.transformer,
    ]
    transformer_sel = select(columns)\
            .where(or_(
                    transformerref.c.attribute_name == 'ship_type',
                    transformerref.c.attribute_name == 'pilot',
            ))
    transformer_rows = conn.execute(transformer_sel)
    for division_id, attribute_name, transformer in transformer_rows:
        if attribute_name == 'ship_type':
            colname = 'ship'
            transformer_class = evesrp.transformers.ShipTransformer
        elif attribute_name == 'pilot':
            colname = 'pilot'
            transformer_class = evesrp.transformers.PilotTransformer
        colname += '_transformer'
        transformer = transformer_class(transformer.name, transformer.slug)
        update_stmt = update(division)\
                .where(division.c.id == division_id)\
                .values({
                        colname: transformer
                })
        conn.execute(update_stmt)
    transformer_rows.close()
    # Drop the transformerref table. This is going to be lossy.
    op.drop_table('transformerref')



"""Restrict request.base_payout to be >= 0

Revision ID: 337978f8c75
Revises: c1fc69b629
Create Date: 2014-06-18 14:04:52.963890

"""

# revision identifiers, used by Alembic.
revision = '337978f8c75'
down_revision = 'c1fc69b629'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column


request = table('request',
        column('id', sa.Integer),
        column('base_payout', sa.Float),
)


def upgrade():
    conn = op.get_bind()
    negative_base_payout_id_sel = select([request.c.id])\
            .where(request.c.base_payout < 0.0)
    negative_ids = conn.execute(negative_base_payout_id_sel)
    for result_row in negative_ids:
        negative_id = result_row[0]
        update_stmt = update(request)\
                .where(request.c.id == negative_id)\
                .values({
                        'base_payout': 0.0,
                })
        conn.execute(update_stmt)
    negative_ids.close()


def downgrade():
    # This is a lossy upgrade, no downgrading possible
    pass



from __future__ import absolute_import
import re

from flask import redirect, url_for, current_app
from flask.ext.login import current_user
import flask.ext.login as flask_login
from flask.ext.wtf import Form
from wtforms.fields import SubmitField, HiddenField
from ..util import DeclEnum, classproperty, ensure_unicode


class AuthForm(Form):
    submit = SubmitField(u'Login')


class AuthMethod(object):
    def __init__(self, admins=None, name=u'Base Authentication', **kwargs):
        if admins is None:
            self.admins = []
        else:
            self.admins = admins
        self.name = ensure_unicode(name)

    def form(self):
        """Return a form class to login with."""
        return AuthForm

    def login(self, form):
        """Process a validated login form.

        You must return a valid response object.
        """
        pass

    def list_groups(self, user=None):
        pass

    def view(self):
        """Optional method for providing secondary views.

        :py:func:`evesrp.views.login.auth_method_login` is configured to allow
        both GET and POST requests, and will call this method as soon as it is
        known which auth method is meant to be called. The path for this view
        is ``/login/self.safe_name/``, and can be generated with
        ``url_for('login.auth_method_login', auth_method=self.safe_name)``.

        The default implementation redirects to the main login view.
        """
        return redirect(url_for('login.login'))

    @staticmethod
    def login_user(user):
        """Signal to the authentication systems that a new user has logged in.

        Handles sending the :py:data:`flask.ext.principal.identity_changed`
        signal and calling :py:func:`flask.ext.login.login_user` for you.

        :param user: The user that has been authenticated and is logging in.
        :type user: :py:class:`~models.User`
        """
        flask_login.login_user(user)

    @property
    def safe_name(self):
        """Normalizes a string to be a valid Python identifier (along with a few
        other things).

        Specifically, all letters are lower cased, only ASCII characters, and
        whitespace replaced by underscores.

        :returns: The normalized string.
        :rtype str:
        """
        # Turn 'fancy' characters into '?'s
        ascii_rep = self.name.encode('ascii', 'replace').decode('utf-8')
        # Whitespace and '?' to underscores
        no_space = re.sub(r'[\s\?]', u'_', ascii_rep)
        lowered = no_space.lower()
        return lowered


class PermissionType(DeclEnum):
    """Enumerated type for the typers of permissions available. """

    #: Permission allowing submission of :py:class:`~.Request`\s to a
    #: :py:class:`~.Division`.
    submit = u'submit', u'Submitter'

    #: Permission for reviewers of requests in a :py:class:`~.Division`.
    review = u'review', u'Reviewer'

    #: Permission for payers in a :py:class:`~.Division`.
    pay = u'pay', u'Payer'

    #: :py:class:`~.Division`\-level administrator permission
    admin = u'admin', u'Administrator'

    @classproperty
    def elevated(cls):
        return frozenset((cls.review, cls.pay, cls.admin))

    @classproperty
    def all(cls):
        return frozenset((cls.submit, cls.review, cls.pay, cls.admin))



from __future__ import absolute_import
import locale
import os
import requests
from flask import Flask, current_app, g
from flask.ext.sqlalchemy import SQLAlchemy
from flask.ext.wtf.csrf import CsrfProtect


db = SQLAlchemy()


from .util import DB_STATS, AcceptRequest


__version__ = u'0.9.1'


requests_session = requests.Session()


# Set default locale
locale.setlocale(locale.LC_ALL, '')


csrf = CsrfProtect()


# Ensure models are declared
from . import models
from .auth import models as auth_models


def create_app(config=None, **kwargs):
    app = Flask('evesrp', **kwargs)
    app.request_class = AcceptRequest
    app.config.from_object('evesrp.default_config')
    if config is not None:
        app.config.from_pyfile(config)
    if app.config['SECRET_KEY'] is None and 'SECRET_KEY' in os.environ:
        app.config['SECRET_KEY'] = os.environ['SECRET_KEY']

    # Register SQLAlchemy monitoring before the DB is connected
    app.before_request(sqlalchemy_before)

    db.init_app(app)

    from .views.login import login_manager
    login_manager.init_app(app)

    before_csrf = list(app.before_request_funcs[None])
    csrf.init_app(app)
    # Remove the context processor that checks CSRF values. All it is used for
    # is the template function.
    app.before_request_funcs[None] = before_csrf

    from .views import index, error_page, divisions, login, requests, api
    app.add_url_rule(rule=u'/', view_func=index)
    for error_code in (400, 403, 404, 500):
        app.register_error_handler(error_code, error_page)
    app.register_blueprint(divisions.blueprint, url_prefix='/division')
    app.register_blueprint(login.blueprint)
    app.register_blueprint(requests.blueprint, url_prefix='/request')
    app.register_blueprint(api.api, url_prefix='/api')
    app.register_blueprint(api.filters, url_prefix='/api/filter')

    from .views import request_count
    app.add_template_global(request_count)

    from .json import SRPEncoder
    app.json_encoder=SRPEncoder

    app.before_first_request(_copy_config_to_authmethods)
    app.before_first_request(_config_requests_session)
    app.before_first_request(_config_killmails)
    app.before_first_request(_copy_url_converter_config)

    # Configure the Jinja context
    # Inject variables into the context
    from .auth import PermissionType
    @app.context_processor
    def inject_enums():
        return {
            'ActionType': models.ActionType,
            'PermissionType': PermissionType,
            'app_version': __version__,
            'site_name': app.config['SRP_SITE_NAME'],
        }
    # Auto-trim whitespace
    app.jinja_env.trim_blocks = True
    app.jinja_env.lstrip_blocks = True

    return app


# SQLAlchemy performance logging
def sqlalchemy_before():
    DB_STATS.clear()
    g.DB_STATS = DB_STATS


# Auth setup
def _copy_config_to_authmethods():
    current_app.auth_methods = current_app.config['SRP_AUTH_METHODS']


# Request detail URL setup
def _copy_url_converter_config():
    url_transformers = {}
    for config_key, config_value in current_app.config.items():
        if config_value is None:
            continue
        index = config_key.rfind('_URL_TRANSFORMERS')
        if not config_key.endswith('_URL_TRANSFORMERS')\
                or not config_key.startswith('SRP_'):
            continue
        attribute = config_key.replace('SRP_', '', 1)
        attribute = attribute.replace('_URL_TRANSFORMERS', '', 1)
        attribute = attribute.lower()
        url_transformers[attribute] = {t.name:t for t in config_value}
    current_app.url_transformers = url_transformers
    # temporary compatibility
    current_app.ship_urls = url_transformers.get('ship_type', None)
    current_app.pilot_urls = url_transformers.get('pilot', None)


# Requests session setup
def _config_requests_session():
    try:
        ua_string = current_app.config['SRP_USER_AGENT_STRING']
    except KeyError as outer_exc:
        try:
            ua_string = 'EVE-SRP/{version} ({email})'.format(
                    email=current_app.config['SRP_USER_AGENT_EMAIL'],
                    version=__version__)
        except KeyError as inner_exc:
            raise inner_exc
    requests_session.headers.update({'User-Agent': ua_string})
    current_app.user_agent = ua_string


# Killmail verification
def _config_killmails():
    current_app.killmail_sources = current_app.config['SRP_KILLMAIL_SOURCES']



from __future__ import unicode_literals
from tempfile import mkdtemp
from urllib.parse import urlparse
from os import path
import bz2
import requests
from sqlalchemy import create_engine, Table, MetaData
from sqlalchemy.sql import select


meta = MetaData()


def download_to_temp(url, dest_path=None):
    if dest_path is None:
        dest_path = mkdtemp()
    parsed_url = urlparse(url)
    filename = path.basename(parsed_url.path)
    dest = path.join(dest_path, filename)
    ext = path.splitext(filename)[1]
    response = requests.get(url, stream=True)
    bunzip = False
    if ext in ('.bz2', '.bz'):
        decompressor = bz2.BZ2Decompressor()
        bunzip = True
    with open(dest, 'wb') as f:
        for chunk in response.iter_content(chunk_size=512):
            if chunk:
                if bunzip:
                    inflated = decompressor.decompress(chunk)
                    if inflated:
                        f.write(inflated)
                else:
                    f.write(chunk)
    return dest


def bind_engine(sqlite_path):
    engine = create_engine('sqlite:///' + sqlite_path)
    meta.bind = engine
    meta.reflect()
    return engine


def get_systems(engine):
    systemNames = {}
    sel = select([meta.tables['mapSystems'].c.solarSystemID,
            meta.tables['mapSystems'].c.solarSystemName])
    result = engine.execute(sel)
    for row in result:
        systemNames[row[0]] = row[1]
    result.close()
    return systemNames


def get_constellations(engine):
    constellationNames = {}
    sel = select([meta.tables['mapConstellations'].c.constellationID,
            meta.tables['mapConstellations'].c.constellationName])
    result = engine.execute(sel)
    for row in result:
        constellationNames[row[0]] = row[1]
    result.close()
    return constellationNames


def get_regions(engine):
    regionNames = {}
    sel = select([meta.tables['mapRegions'].c.regionID,
            meta.tables['mapRegions'].c.regionName])
    result = engine.execute(sel)
    for row in result:
        regionNames[row[0]] = row[1]
    result.close()
    return regionNames


def get_system_constellations(engine, constellation_names):
    systemConstellations = {}
    sel = select([meta.tables['mapSystems'].c.solarSystemName,
            meta.tables['mapSystems'].c.constellationID])
    result = engine.execute(sel)
    for row in result:
        systemConstellations[row[0]] = constellation_names[row[1]]
    result.close()
    return systemConstellations


def get_constellation_regions(engine, region_names):
    constellationRegions = {}
    sel = select([meta.tables['mapConstellations'].c.constellationName,
            meta.tables['mapConstellations'].c.regionID])
    result = engine.execute(sel)
    for row in result:
        constellationRegions[row[0]] = region_names[row[1]]
    result.close()
    return constellationRegions


def get_ships(engine):
    # Get the Ship category ID
    cat_sel = select([meta.tables['invCategories'].c.categoryID]).\
            where(meta.tables['invCategories'].c.categoryName == 'Ship').limit(1)
    result = engine.execute(cat_sel)
    cat_id = result.fetchone()[0]
    result.close()

    # Get groups in this category
    group_sel = select([meta.tables['invGroups'].c.groupID]).where(
            meta.tables['invGroups'].c.categoryID == cat_id)
    group_result = engine.execute(group_sel)

    # get typeIDs and names
    names = {}
    for group_row in group_result:
        group_id = group_row[0]
        type_sel = select([meta.tables['invTypes'].c.typeID,
                meta.tables['invTypes'].c.typeName]).\
                where(meta.tables['invTypes'].c.groupID == group_id)
        type_result = engine.execute(type_sel)
        for row in type_result:
            names[row[0]] = row[1]
        type_result.close()
    group_result.close()

    return names


def update_static_data(data_dir=None):
    db_path = download_to_temp('https://www.fuzzwork.co.uk/dump/'
                               'sqlite-latest.sqlite.bz2')
    engine = bind_engine(db_path)
    # stuff...



from __future__ import absolute_import
from __future__ import unicode_literals
from flask import redirect, url_for, render_template, make_response, request
from flask.ext.login import login_required, current_user
from .. import db
from ..models import Request, ActionType
from ..auth import PermissionType
from ..auth.models import Permission, Division
from ..util import jsonify


@login_required
def index():
    """The index page for EVE-SRP."""
    return redirect(url_for('requests.personal_requests'))


def error_page(error):
    """View function for displaying error pages."""
    # Try to get some meaningful bits of information about the error
    # HTTPExceptions (raised by abort()) have code and description attributes.
    # Try to prefer those rich values over more generic information.
    code = error.code if hasattr(error, 'code') else 500
    if hasattr(error, 'description'):
        description = error.description
    else:
        description = str(error)
    name = error.name if hasattr(error, 'name') else u'Application Error'
    # Give the error information in a machine readable format for APIs
    if request.is_json or request.is_xhr:
        response_content = jsonify(description=error.description,
                code=code)
    elif request.is_xml:
        pass
    else:
        response_content = render_template('error.html', code=code,
                description=description, name=name)
    # Give a default response code for generic exceptions
    return make_response(response_content, code)


def request_count(permission, statuses=None):
    """Function intended for counting the number of requests for Jinja
    templates.
    """
    if statuses is None:
        if permission == PermissionType.review:
            statuses = (ActionType.evaluating,)
        elif permission == PermissionType.pay:
            statuses = (ActionType.approved,)
        elif permission == PermissionType.submit:
            statuses = (ActionType.incomplete,)
        else:
            return 0
    elif statuses in ActionType.statuses:
        statuses = (statuses,)
    permissions = current_user.permissions.\
            filter(Permission.permission==permission).\
            subquery()
    divisions = db.session.query(Division.id).\
            join(permissions).\
            subquery()
    requests = db.session.query(db.func.count(db.distinct(Request.id))).\
            join(divisions).\
            filter(Request.status.in_(statuses))
    if permission == PermissionType.submit:
        requests = requests.filter(Request.submitter==current_user)
    return requests.one()[0]



"""Unicode helper functions."""


import six


def unistr(klass):
    """Class decorator that aids in model compatibility between Python 2 and 3.

    The decorated class needs to define a :py:func:`__unicode__` function (even
    for Python 3). This decorator adds a :py:func:`__str__` function to the
    class that takes :py:func:`__unicode__`\'s return value and encodes it as
    UTF-8 on Python 2 or returns it unmodified on Python 3.
    """
    def __str__(self):
        if six.PY2:
            return self.__unicode__().encode('utf-8')
        else:
            return self.__unicode__()

    klass.__str__ = __str__
    return klass


def ensure_unicode(val):
    """Ensure that a a given value is a unicode string.

    If ``val`` is a :py:data:`six.binary_type <binary type>`, assume it's UTF-8
    and decode it.
    :param val: A :py:data:`six.binary_type <binary>` or
        :py:data:`six.text_type <text>` type.
    :rtype: :py:class:`unicode` on Python 2, :py:class:`str` on Python 3
    """
    if isinstance(val, six.binary_type):
        return val.decode('utf-8')
    return val



from __future__ import absolute_import

"""Originally taken from http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/
on 23 May 2014. Specifically, this is a modified version of
http://techspot.zzzeek.org/files/2011/decl_enum.py
"""

import six
from sqlalchemy.types import SchemaType
import re
from .unistr import unistr, ensure_unicode
from .. import db


# NOTE: When adding Py2 support, make sure to set the metaclasses appropriately


@unistr
class EnumSymbol(object):
    """Define a fixed symbol tied to a parent class."""

    def __init__(self, cls_, name, value, description):
        self.cls_ = cls_
        self.name = name
        self.value = value
        self.description = description

    def __reduce__(self):
        """Allow unpickling to return the symbol 
        linked to the DeclEnum class."""
        return getattr, (self.cls_, self.name)

    def __iter__(self):
        return iter([self.value, self.description])

    def __repr__(self):
        return u"<%s>" % self.name

    def __unicode__(self):
        return self.description


class EnumMeta(type):
    """Generate new DeclEnum classes."""

    def __init__(cls, classname, bases, dict_):
        cls._reg = reg = cls._reg.copy()
        for k, v in six.iteritems(dict_):
            if isinstance(v, tuple):
                unicoded = []
                for tup_v in v:
                    unicoded.append(ensure_unicode(tup_v))
                sym = reg[unicoded[0]] = EnumSymbol(cls, k, *unicoded)
                setattr(cls, k, sym)
        return type.__init__(cls, classname, bases, dict_)

    def __iter__(cls):
        return six.itervalues(cls._reg)


@six.add_metaclass(EnumMeta)
class DeclEnum(object):
    """Declarative enumeration."""

    _reg = {}

    @classmethod
    def from_string(cls, value):
        try:
            return cls._reg[ensure_unicode(value)]
        except KeyError:
            raise ValueError(
                    u"Invalid value for %r: %r" % 
                    (cls.__name__, ensure_unicode(value))
                )

    @classmethod
    def values(cls):
        return six.iterkeys(cls._reg)

    @classmethod
    def db_type(cls):
        return DeclEnumType(cls)


class DeclEnumType(SchemaType, db.TypeDecorator):
    def __init__(self, enum):
        self.enum = enum
        self.impl = db.Enum(
                        *(list(enum.values())),
                        convert_unicode=True,
                        name=u"ck%s" % re.sub(
                                    '([A-Z])', 
                                    lambda m:"_" + m.group(1).lower(), 
                                    enum.__name__)
                    )

    def _set_table(self, table, column):
        self.impl._set_table(table, column)

    def copy(self):
        return DeclEnumType(self.enum)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum.from_string(value.strip())



"""Split modifier types out into classes.

Revision ID: 2976d59f286
Revises: 4198a248c8a
Create Date: 2014-06-18 15:19:55.611649

"""

# revision identifiers, used by Alembic.
revision = '2976d59f286'
down_revision = '4198a248c8a'

from decimal import Decimal
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import update, select, table, column


modifier = table('modifier',
        column('id', sa.Integer),
        column('value', sa.Float),
        column('type_', sa.Enum('absolute', 'percentage',
                name='modifier_type')),
        column('_type', sa.String(length=20)),
)


abs_table = table('absolute_modifier',
        column('id', sa.Integer),
        column('value', sa.Numeric(precision=15, scale=2)))


rel_table = table('relative_modifier',
        column('id', sa.Integer),
        column('value', sa.Float))


def upgrade():
    # Add discriminator column
    op.add_column('modifier', sa.Column('_type', sa.String(length=20)))
    # Create new subclass tables
    op.create_table('absolute_modifier',
            sa.Column('id', sa.Integer,
                    sa.ForeignKey('modifier.id'),
                    primary_key=True),
            sa.Column('value', sa.Numeric(precision=15, scale=2),
                    nullable=False, server_default='0.0'))
    op.create_table('relative_modifier',
            sa.Column('id', sa.Integer, sa.ForeignKey('modifier.id'),
                    primary_key=True),
            sa.Column('value', sa.Float, nullable=False, server_default='0.0'))
    # Add new entries to the subclass tables for each modifier
    conn = op.get_bind()
    modifier_sel = select([modifier.c.id, modifier.c.value, modifier.c.type_])
    modifiers = conn.execute(modifier_sel)
    absolutes = []
    relatives = []
    for modifier_id, modifier_value, modifier_type in modifiers:
        if modifier_type == 'absolute':
            discriminator = 'AbsoluteModifier'
            absolutes.append({
                    'id': modifier_id,
                    'value': Decimal.from_float(modifier_value) * 1000000,
            })
        elif modifier_type == 'percentage':
            discriminator = 'RelativeModifier'
            relatives.append({
                    'id': modifier_id,
                    'value': modifier_value / 100,
            })
        update_stmt = update(modifier)\
                .where(modifier.c.id == modifier_id)\
                .values({
                        '_type': discriminator,
                })
        conn.execute(update_stmt)
    modifiers.close()
    op.bulk_insert(abs_table, absolutes)
    op.bulk_insert(rel_table, relatives)
    # Drop the old value and type_ columns from modifier
    op.drop_column('modifier', 'value')
    op.drop_column('modifier', 'type_')
    # Add the not-null constraint to the _type column
    op.alter_column('modifier',
            column_name='_type',
            nullable=True,
            existing_type=sa.String(length=20),
    )


def downgrade():
    # Add type_ and value columns back
    op.add_column('modifier',
            sa.Column('type_', sa.Enum('absolute', 'percentage',
                    name='modifier_type')))
    op.add_column('modifier', sa.Column('value', sa.Float, nullable=True))
    # populate type_ and value columns with data from the subclass tables
    abs_select = select([abs_table.c.id, abs_table.c.value])
    rel_select = select([rel_table.c.id, rel_table.c.value])
    conn = op.get_bind()
    for select_stmt in (abs_select, rel_select):
        modifier_rows = conn.execute(select_stmt)
        for modifier_id, modifier_value in modifier_rows:
            if select_stmt == abs_select:
                modifier_value = float(modifier_value / 1000000)
                type_ = 'absolute'
            else:
                type_ = 'percentage'
            update_stmt = update(modifier)\
                    .where(modifier.c.id == modifier_id)\
                    .values({
                        'value': modifier_value,
                        'type_': type_
                    })
            conn.execute(update_stmt)
        modifier_rows.close()
    # Drop the old _type column and the subclass tables
    op.drop_column('modifier', '_type')
    op.drop_table('absolute_modifier')
    op.drop_table('relative_modifier')
    # Add not-null constraint back to type_
    op.alter_column('modifier',
            column_name='type_',
            nullable=False,
            existing_type=sa.Enum('absolute', 'percentage',
                    name='modifier_type'))
